#include "NVSExplorer.hpp"

// C++ system level
#include <set>
#include <cmath>
#include <limits>      // mix/max integer values; used for parsing network data
#include <vector>
#include <cstring>     // memset, memcpy, strncmp
#include <cstdlib>     // std::strtof and such…
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_err.h"
// #include "driver/uart.h"

// project specific
#include <Types.hpp>
#include <SpiderLib/Util.hpp>

// qthing stuff
#include <qthing>
#include <qthing/mqtt_common.hpp>

// misc
#include <nlohmann/json.hpp>


using json = nlohmann::json;




namespace SiliconTorch {

  namespace NVSExplorer {


    const char* TAG = "NVSExplorer";


    NVSExplorer::NVSExplorer() {

      ESP_LOGI(TAG, "Starting NVSExplorer[📡]");

      nvs_flash_init();

      registerLambdas0();
      registerLambdas1();

    }  // end NVSExplorer()-ctor



    void NVSExplorer::registerLambdas0() {

      u16 deviceTopicLEN = requestTopic("").length();

      auto wrapMessageHandler = [deviceTopicLEN](std::function<void(std::vector<str>&, qthing::multipart_message_t)> handler) {
        return [handler, deviceTopicLEN](qthing::multipart_message_t msg) {
          if (msg.length == msg.total_length && msg.offset == 0) {

            str localTopic(msg.topic.c_str() + deviceTopicLEN);
            std::vector<str> topicParts = SpiderLib::Util::split(localTopic, '/');

            handler(topicParts, msg);
          } else {
            ESP_LOGW(TAG, "Multipart messages are not supported atm.  topic[ %s ]", msg.topic.c_str());
          }
        };
      };



      // TODO: publish error if NS not found…?!
      qthing::add_binary_message_callback(requestTopic("ls/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        
        str delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = str(buffer);
        }

        auto keys = listKeys(topic[1]);
        str msgOut = SpiderLib::Util::join(keys, delimiter);

        qthing::publish_message(responseTopic(topic[1]), msgOut);

        if (buffer != NULL) delete buffer;
      } ));



      qthing::add_binary_message_callback(requestTopic("ls"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        
        str delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = str(buffer);
        }

        auto nameSpaces = listNamespaces();
        str msgOut = SpiderLib::Util::join(nameSpaces, delimiter);

        qthing::publish_message(responseTopic("ls"), msgOut);

        if (buffer != NULL) delete buffer;
      } ));



      qthing::add_binary_message_callback(requestTopic("rm/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        this->removeKey(topic[1], topic[2]);

        // publish error…?
      } ));



      qthing::add_binary_message_callback(requestTopic("get_float/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        float value = this->getFloat(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_float", topic[1], topic[2]), str(buffer));
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving float[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      } ));



      qthing::add_binary_message_callback(requestTopic("get_double/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        double value = this->getDouble(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_double", topic[1], topic[2]), str(buffer));
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving double[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      } ));



      qthing::add_binary_message_callback(requestTopic("set_float/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        float value = std::strtof(buffer, NULL);

        if ( this->setFloat(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to float[ %f ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to float[ %f ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);

        // publish error…?
      } ));



      qthing::add_binary_message_callback(requestTopic("set_double/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        double value = std::strtod(buffer, NULL);

        if ( this->setDouble(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set namespace[ %s ]  key[ %s ] to double[ %f ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting namespace[ %s ]  key[ %s ] to double[ %f ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);

        // publish error…?
      } ));



      qthing::add_binary_message_callback(requestTopic("get_u8/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u8 value = 0;
          if ( nvs_get_u8(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u8", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u8[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_u16/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u16 value = 0;
          if ( nvs_get_u16(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u16", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u16[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_u32/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u32 value = 0;
          if ( nvs_get_u32(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u32", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u32[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_u64/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u64 value = 0;
          if ( nvs_get_u64(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%lld", value);
            qthing::publish_message(responseTopic("get_u64", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u64[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_i8/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i8 value = 0;
          if ( nvs_get_i8(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i8", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i8[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_i16/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i16 value = 0;
          if ( nvs_get_i16(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i16", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i16[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_i32/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i32 value = 0;
          if ( nvs_get_i32(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i32", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i32[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_i64/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i64 value = 0;
          if ( nvs_get_i64(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%lld", value);
            qthing::publish_message(responseTopic("get_i64", topic[1], topic[2]), str(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i64[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      } ));


    }



    void NVSExplorer::registerLambdas1() {

      u16 deviceTopicLEN = requestTopic("").length();

      auto wrapMessageHandler = [deviceTopicLEN](std::function<void(std::vector<str>&, qthing::multipart_message_t)> handler) {
        return [handler, deviceTopicLEN](qthing::multipart_message_t msg) {
          if (msg.length == msg.total_length && msg.offset == 0) {

            str localTopic(msg.topic.c_str() + deviceTopicLEN);
            std::vector<str> topicParts = SpiderLib::Util::split(localTopic, '/');

            handler(topicParts, msg);
          } else {
            ESP_LOGW(TAG, "Multipart messages are not supported atm.  topic[ %s ]", msg.topic.c_str());
          }
        };
      };



      qthing::add_binary_message_callback(requestTopic("set_u8/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u8 u8_max = std::numeric_limits<u8>::max();
        if (value > u8_max) value = u8_max;

        if ( this->setU8(topic[1], topic[2], static_cast<u8>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u8[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u8[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_u16/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u16 u16_max = std::numeric_limits<u16>::max();
        if (value > u16_max) value = u16_max;

        if ( this->setU16(topic[1], topic[2], static_cast<u16>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u16[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u16[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_u32/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u32 u32_max = std::numeric_limits<u32>::max();
        if (value > u32_max) value = u32_max;

        if ( this->setU32(topic[1], topic[2], static_cast<u32>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u32[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u32[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_u64/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        if ( this->setU64(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u64[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u64[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_i8/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i8 i8_max = std::numeric_limits<i8>::max();
        const i8 i8_min = std::numeric_limits<i8>::lowest();

        if (value > i8_max) value = i8_max;
        if (value < i8_min) value = i8_min;

        if ( this->setI8(topic[1], topic[2], static_cast<i8>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i8[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i8[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_i16/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i16 i16_max = std::numeric_limits<i16>::max();
        const i16 i16_min = std::numeric_limits<i16>::lowest();

        if (value > i16_max) value = i16_max;
        if (value < i16_min) value = i16_min;

        if ( this->setI16(topic[1], topic[2], static_cast<i16>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i16[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i16[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_i32/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i32 i32_max = std::numeric_limits<i32>::max();
        const i32 i32_min = std::numeric_limits<i32>::lowest();

        if (value > i32_max) value = i32_max;
        if (value < i32_min) value = i32_min;

        if ( this->setI32(topic[1], topic[2], static_cast<i32>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i32[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i32[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("set_i64/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        if ( this->setI64(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i64[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i64[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      } ));



      qthing::add_binary_message_callback(requestTopic("get_strLen/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        size_t testValue = 1337000;  // way too big for NVS
        size_t length = getStringLength(topic[1], topic[2], testValue);

        if (length < testValue) {  // key exists and is of type string

          char buffer[32];
          std::snprintf(&buffer[0], 32, "%d", length);

          str value(&buffer[0]);
          qthing::publish_message(responseTopic("get_strLen", topic[1], topic[2]), value);
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving |str|[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("get_str/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        size_t testValue = 1337000;  // way too big for NVS

        if ( getStringLength(topic[1], topic[2], testValue) < testValue ) {  // key exists and is of type string

          str value = getString(topic[1], topic[2]);

          qthing::publish_message(responseTopic("get_str", topic[1], topic[2]), value);
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving str[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      } ));


      qthing::add_binary_message_callback(requestTopic("set_str/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);
        str value(buffer);

        if ( this->setString(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to str[ %s ]",             topic[1].c_str(), topic[2].c_str(), value.c_str());  // TODO: do we REALLY wanna log the payloads?!
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to str[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str(), value.c_str());  // TODO: do we REALLY wanna log the payloads?!
        // publish error…?
      } ));


      qthing::add_binary_message_callback(requestTopic("get_type/+/+"), wrapMessageHandler(
        [&](std::vector<str>& topic, qthing::multipart_message_t msg) {
        qthing::publish_message(responseTopic("get_type", topic[1], topic[2]), getTypeStr(topic[1], topic[2]));
      } ));


      qthing::add_message_callback(requestTopic("export_json"), [&](str message) {
        publishJSON( message.compare("pretty") == 0 );
      });


    }



    u8 NVSExplorer::getU8(const str& nameSpace, const str& key, u8 defaultValue) {

      u8 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u8(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u16 NVSExplorer::getU16(const str& nameSpace, const str& key, u16 defaultValue) {

      u16 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u16(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u32 NVSExplorer::getU32(const str& nameSpace, const str& key, u32 defaultValue) {

      u32 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u32(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u64 NVSExplorer::getU64(const str& nameSpace, const str& key, u64 defaultValue) {

      u64 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u64(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i8 NVSExplorer::getI8(const str& nameSpace, const str& key, i8 defaultValue) {

      i8 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i8(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i16 NVSExplorer::getI16(const str& nameSpace, const str& key, i16 defaultValue) {

      i16 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i16(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i32 NVSExplorer::getI32(const str& nameSpace, const str& key, i32 defaultValue) {

      i32 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i32(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i64 NVSExplorer::getI64(const str& nameSpace, const str& key, i64 defaultValue) {

      i64 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i64(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }



    bool NVSExplorer::setU8(const str& nameSpace, const str& key, u8 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u8(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU16(const str& nameSpace, const str& key, u16 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u16(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU32(const str& nameSpace, const str& key, u32 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u32(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU64(const str& nameSpace, const str& key, u64 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u64(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI8(const str& nameSpace, const str& key, i8 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i8(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI16(const str& nameSpace, const str& key, i16 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i16(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI32(const str& nameSpace, const str& key, i32 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i32(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI64(const str& nameSpace, const str& key, i64 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i64(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }



    float NVSExplorer::getFloat(const str& nameSpace, const str& key) {

      float value  = std::nanf("1337");
      u32*  valptr = reinterpret_cast<u32*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u32(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }

    double NVSExplorer::getDouble(const str& nameSpace, const str& key) {

      double value  = std::nan("1337");
      u64*   valptr = reinterpret_cast<u64*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u64(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }


    bool NVSExplorer::setFloat(const str& nameSpace, const str& key, float value) {

      bool success = true;
      u32  intval  = reinterpret_cast<u32*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u32(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setDouble(const str& nameSpace, const str& key, double value) {

      bool success = true;
      u64  intval  = reinterpret_cast<u64*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u64(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }


    size_t NVSExplorer::getStringLength(const str& nameSpace, const str& key, size_t defaultValue) {

      size_t value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_str(nvs, key.c_str(), NULL, &value);
           nvs_close(nvs);
      }

      return value;
    }


    str NVSExplorer::getString(const str& nameSpace, const str& key, const str& defaultValue) {

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

        size_t size;
        if ( nvs_get_str(nvs, key.c_str(), NULL, &size) == ESP_OK ) {

          char buffer[size + 1];
          std::memset(&buffer[0], 0x00, size + 1);

          nvs_get_str(nvs, key.c_str(), &buffer[0], &size);

          nvs_close(nvs);

          str out(&buffer[0]);
          return out;
        }
      }

      return defaultValue;
    }


    bool NVSExplorer::setString(const str& nameSpace, const str& key, const str& value) {

      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_str(nvs, key.c_str(), value.c_str()) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }



    bool NVSExplorer::removeKey(const str& nameSpace, const str& key) {
      nvs_handle_t _nvs;
      ESP_LOGW(TAG, "removeKey() -> %s / %s", nameSpace.c_str(), key.c_str());
      ESP_LOGW(TAG, "open() -> %s", esp_err_to_name( nvs_open(nameSpace.c_str(), NVS_READWRITE, &_nvs) ));
      ESP_LOGW(TAG, "erase() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
      ESP_LOGW(TAG, "commit() -> %s", esp_err_to_name( nvs_commit(_nvs) ));
      ESP_LOGW(TAG, "erase²() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));

      return true;
    }

    std::set<str> NVSExplorer::listKeys(const str& nameSpace) {

      std::set<str> keys;

      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        keys.insert(info.key);
      }

      return keys;
    }

    std::set<str> NVSExplorer::listNamespaces() {

      std::set<str> namespaces;

      nvs_iterator_t it = nvs_entry_find("nvs", NULL, NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        namespaces.insert(info.namespace_name);
      }

      return namespaces;
    }


    i64 NVSExplorer::getSignedInt(const str& nameSpace, const str& key, i64 defaultValue) {

      // Lets first test & evaluate the other function before duplication………

      return -1337LL;
    }


    u64 NVSExplorer::getUnsignedInt(const str& nameSpace, const str& key, u64 defaultValue) {

      u64 value = defaultValue;
      nvs_type_t entryType = NVS_TYPE_ANY;

      // search for the key to get its type
      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t entry;
        nvs_entry_info(it, &entry);

        if (key.compare(entry.key) == 0) {
          entryType = entry.type;
          nvs_release_iterator(it);
          break;
        }

        it = nvs_entry_next(it);
      }

      if (entryType == NVS_TYPE_ANY)  // key not found
        return defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) != ESP_OK )  // namespace not found
        return defaultValue;

      u64* valptr = &value;
      switch (entryType) {
        case NVS_TYPE_U8:  { nvs_get_u8(nvs,  key.c_str(), ( u8*)valptr); break; }
        case NVS_TYPE_I8:  { nvs_get_i8(nvs,  key.c_str(), (  i8*)valptr); break; }
        case NVS_TYPE_U16: { nvs_get_u16(nvs, key.c_str(), (u16*)valptr); break; }
        case NVS_TYPE_I16: { nvs_get_i16(nvs, key.c_str(), ( i16*)valptr); break; }
        case NVS_TYPE_U32: { nvs_get_u32(nvs, key.c_str(), (u32*)valptr); break; }
        case NVS_TYPE_I32: { nvs_get_i32(nvs, key.c_str(), ( i32*)valptr); break; }
        case NVS_TYPE_U64: { nvs_get_u64(nvs, key.c_str(), (u64*)valptr); break; }
        case NVS_TYPE_I64: { nvs_get_i64(nvs, key.c_str(), ( i64*)valptr); break; }
        default: {}
      }

      nvs_close(nvs);
      return value;
    }


    bool NVSExplorer::setSignedInt(const str& nameSpace, const str& key, i64 value, nvs_type_t entryType) {

      // TODO: implementation!

      return false;
    }

    bool NVSExplorer::setUnsignedInt(const str& nameSpace, const str& key, u64 value, nvs_type_t entryType) {

      bool result = false;

      if (entryType == NVS_TYPE_ANY) {

        // check if key exists and grab its type
        nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t entry;
          nvs_entry_info(it, &entry);

          if (key.compare(entry.key) == 0) {
            entryType = entry.type;
            nvs_release_iterator(it);
            break;
          }

          it = nvs_entry_next(it);
        }

        // TODO: further implementation……?

      } else {

      }

      return result;
    }


    /**
      * Converts IDF NVS types to strings
      */
    str NVSExplorer::type2str(nvs_type_t typ) {
      switch (typ) {
        case NVS_TYPE_U8:   return "u8";
        case NVS_TYPE_U16:  return "u16";
        case NVS_TYPE_U32:  return "u32";
        case NVS_TYPE_U64:  return "u64";
        case NVS_TYPE_I8:   return "i8";
        case NVS_TYPE_I16:  return "i16";
        case NVS_TYPE_I32:  return "i32";
        case NVS_TYPE_I64:  return "i64";
        case NVS_TYPE_STR:  return "str";
        case NVS_TYPE_BLOB: return "blob";
        default:            return "any";
      }
    }


    /**
      * Get the type of the entry, NVS_TYPE_ANY if not found
      */
    nvs_type_t NVSExplorer::getType(const str& nameSpace, const str& key) {

      nvs_type_t entryType = NVS_TYPE_ANY;

      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t entry;
        nvs_entry_info(it, &entry);

        if (key.compare(entry.key) == 0) {
          entryType = entry.type;
          nvs_release_iterator(it);
          return entryType;
        }

        it = nvs_entry_next(it);
      }

      return entryType;
    }

    /**
      * Like getType but converts them to string
      * non-existing keys are converted to `void` instead of `any`
      */
    str NVSExplorer::getTypeStr(const str& nameSpace, const str& key) {

      bool entryFound = false;
      nvs_type_t entryType = NVS_TYPE_ANY;

      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t entry;
        nvs_entry_info(it, &entry);

        if (key.compare(entry.key) == 0) {
          entryFound = true;
          entryType = entry.type;
          nvs_release_iterator(it);
          break;
        }

        it = nvs_entry_next(it);
      }

      str typ;

      if (entryFound) typ = type2str(entryType);
      else            typ = "void";

      return typ;
    }



    bool NVSExplorer::copyKey(const str& nameSpace, const str& keyFrom, const str& keyTo) {

      return false;  // TODO: implementation!

    }

    bool NVSExplorer::copyKey(const str& nsFrom, const str& keyFrom, const str& nsTo, const str& keyTo) {

      nvs_type_t entryType = getType(nsFrom, keyFrom);

      if (entryType == NVS_TYPE_ANY)  // key not found
        return false;

      if (entryType == NVS_TYPE_BLOB)  // BLOBs are currently unsupported
        return false;


      // u64 intStorage = 0;
      // str strStorage = "";


      return false;

    }


    bool NVSExplorer::moveKey(const str& nameSpace, const str& keyFrom, const str& keyTo) {

      if ( ! copyKey(nameSpace, keyFrom, keyTo) )
        return false;

      return removeKey(nameSpace, keyFrom);
    }

    bool NVSExplorer::moveKey(const str& nsFrom, const str& keyFrom, const str& nsTo, const str& keyTo) {

      if ( ! copyKey(nsFrom, keyFrom, nsTo, keyTo) )
        return false;

      return removeKey(nsFrom, keyFrom);
    }


    str NVSExplorer::exportJSON(bool prettyPrint) {

      json storage;

      for (const auto& ns : listNamespaces()) {
        json keys;

        nvs_iterator_t it = nvs_entry_find("nvs", ns.c_str(), NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t entry;
          nvs_entry_info(it, &entry);

          json _entry;
          str key(entry.key);
          bool entryOK = true;

          switch (entry.type) {
            case NVS_TYPE_U8:  { _entry["value"] = getU8    (ns, key); break; }
            case NVS_TYPE_U16: { _entry["value"] = getU16   (ns, key); break; }
            case NVS_TYPE_U32: { _entry["value"] = getU32   (ns, key); break; }
            case NVS_TYPE_U64: { _entry["value"] = getU64   (ns, key); break; }
            case NVS_TYPE_I8:  { _entry["value"] = getI8    (ns, key); break; }
            case NVS_TYPE_I16: { _entry["value"] = getI16   (ns, key); break; }
            case NVS_TYPE_I32: { _entry["value"] = getI32   (ns, key); break; }
            case NVS_TYPE_I64: { _entry["value"] = getI64   (ns, key); break; }
            case NVS_TYPE_STR: { _entry["value"] = getString(ns, key); break; }
            // case NVS_TYPE_BLOB: {}  // BLOB type currently unsupported
            default: { entryOK = false; }
          }

          if (entryOK) {
            _entry["type"] = type2str(entry.type);
            keys[entry.key] = _entry;
          }

          it = nvs_entry_next(it);
        }

        if (keys.is_object())
          storage[ns] = keys;
      }


      nvs_stats_t nvs_stats;
      nvs_get_stats("nvs", &nvs_stats);

      json stats;

      stats["used_entries"] = nvs_stats.used_entries;
      stats["free_entries"] = nvs_stats.free_entries;
      stats["total_entries"] = nvs_stats.total_entries;


      json out;

      out["stats"]   = stats;
      out["storage"] = storage;

      return out.dump(prettyPrint ? 2 : -1);
    }

    void NVSExplorer::publishJSON(bool prettyPrint) {
      qthing::publish_message(responseTopic("export_json"), exportJSON(prettyPrint).c_str());
    }


    str NVSExplorer::requestTopic(const str& cmd) const {
      return deviceTopic("NVS", "rq", cmd);
    }

    str NVSExplorer::requestTopic(const str& cmd, const str& nameSpace) const {
      return deviceTopic("NVS", "rq", cmd, nameSpace);
    }

    str NVSExplorer::requestTopic(const str& cmd, const str& nameSpace, const str& key) const {
      return deviceTopic("NVS", "rq", cmd, nameSpace, key);
    }


    str NVSExplorer::responseTopic(const str& cmd) const {
      return deviceTopic("NVS", "rs", cmd);
    }

    str NVSExplorer::responseTopic(const str& cmd, const str& nameSpace) const {
      return deviceTopic("NVS", "rs", cmd, nameSpace);
    }

    str NVSExplorer::responseTopic(const str& cmd, const str& nameSpace, const str& key) const {
      return deviceTopic("NVS", "rs", cmd, nameSpace, key);
    }

  }
}
